"""
$RCSfile: ExternalFile.py,v $
Implements ExternalFile class as a Zope product
ExternalFile encapsulates a file in the filesystem for
use as a Zope object.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/01/17 06:57:34 $'[6:-2]
__version__ = '$Revision: 1.55 $'[10:-2]

#
# Add our base classes.  Derived classes will get this stuff automatically
#
from OFS.SimpleItem import SimpleItem  # The SimpleItem module is a Zope builtin that provides
                                       # cut/paste, management views, webdav/ftp, undo, ownership,
                                       # traversal, acquisition, and persistence
from Products.ZCatalog.CatalogAwareness import CatalogAware # Reqd for ZCatalog support
from OFS.PropertyManager import PropertyManager  # Reqd for proper property mgmt

# Zope builtins
from Globals import DTMLFile, MessageDialog
import Globals               # InitializeClass (security stuff)
import OFS                   # we grab his _property settings
from DateTime.DateTime import DateTime
from webdav.common import rfc1123_date
from AccessControl import ClassSecurityInfo

# Python builtins
from cStringIO import StringIO
import errno, mimetypes, os, os.path, socket, string, sys, tempfile, types

################################################################
# Container Methods
################################################################

# globals() is standard python for passing the current namespace
manage_addForm                = DTMLFile('dtml/create', globals())
create_externalfile_formpart  = DTMLFile('dtml/create_externalfile_formpart', globals())
create_standard_formpart      = DTMLFile('dtml/create_standard_formpart', globals())
confirm_create_form           = DTMLFile('dtml/confirm_create_form', globals())
conflict_create_form          = DTMLFile('dtml/conflict_create_form', globals())

#
# Map to determine how to display files based on their mime type.
#
# ascii:  in an HTML <textarea>
# image:  in an HTML <image>
# binary: download only
#
# notfound = binary
#
default_content_type = 'application/octet-stream'

mime_type_map     = {
    'image/gif'                     : 'image',
    'image/jpeg'                    : 'image',
    'image/x-portable-bitmap'       : 'image',
    'image/png'                     : 'image',
    'image/tiff'                    : 'image',
    'application/pdf'               : 'binary',
    'application/zip'               : 'binary',
    'application/x-tar'             : 'binary',
    'application/msword'            : 'binary',
    'application/vnd.ms-excel'      : 'binary',
    'application/vnd.ms-powerpoint' : 'binary',
    'application/xml'               : 'ascii',
    'text/plain'                    : 'ascii',
    'text/html'                     : 'ascii',
    'text/dtd'                      : 'ascii',
    'text/stx'                      : 'ascii',
    'text/xml'                      : 'ascii',
    'text/xsl'                      : 'ascii',
    'text/css'                      : 'ascii',
    }

addl_fileext_to_mime_types_map = {
    '.css'  : 'text/css',
    '.dtd'  : 'text/dtd',  # XML DTD files
    '.ent'  : 'text/dtd',  # XML DTD Entity files
    '.stx'  : 'text/stx',  # Structured Text files
    '.dtml' : 'text/html', # DTML markup
    '.doc'  : 'application/msword',  
    '.xls'  : 'application/vnd.ms-excel',
    '.ppt'  : 'application/vnd.ms-powerpoint',
    }

def _copy(infile, outfile):
    """

    Read binary data from infile and write it to outfile. infile and
    outfile my be strings, in which case a file with that name is
    opened, or filehandles, in which case they are accessed directly.

    """
    if type(infile) is types.StringType: 
        instream = open(infile, 'rb')
        close_in = 1
    else:
        close_in = 0
        instream = infile
        
    if type(outfile) is types.StringType: 
        outstream = open(outfile, 'wb')
        close_out = 1
    else:
        close_out = 0
        outstream = outfile
    try:
        blocksize = 2<<16
        block = instream.read(blocksize)
        outstream.write(block)
        while len(block)==blocksize:
            block = instream.read(blocksize)
            outstream.write(block)
        instream.seek(0)
    finally:
        if close_in: instream.close()
        if close_out: outstream.close()
                    
def confirm_create_action(self, file_id, file_title,
                          file_description, basedir,
                          target_filepath,
                          temp_filename, upload_filename,
                          REQUEST, RESPONSE):
    """
    
    Create a new instance of ExternalFile pointing to target filepath
    on server.  If user has uploaded a file (upload_file is not None),
    overwrite the contents of the server file with the uploaded file
    contents.
    
    """
    target_is_existing_file = os.path.exists(basedir + target_filepath)

    create_msg = ""
    if upload_filename:
        try:
            create_msg = self.create_file(temp_filename, basedir, target_filepath)
        except IOError, value:
            message = """Target file %s is not writable
            on server %s (%s).\n""" % (
                basedir + target_filepath, self.getHostname(), value )
            message.replace('\n','<br/>')
            return MessageDialog(title   = 'ERROR',
                                 message = message,
                                 action  = 'create')
        if target_is_existing_file:
            message = """Object instance %s successfully created,
            pointing to file %s on %s.
            Contents of %s:%s
            were replaced by contents of uploaded file %s.
            %s\n""" % (
                file_id, basedir + target_filepath, self.getHostname(),
                self.getHostname(), basedir + target_filepath,
                upload_filename, create_msg)
        else:
            message = """Object instance %s successfully created,
            pointing to newly-created file %s on %s,
            with contents copied from uploaded file %s.
            %s\n""" % (
                file_id, basedir + target_filepath, self.getHostname(),
                upload_filename, create_msg )
    else:
        if not target_is_existing_file:
            try:
                create_msg = self.create_file(None, basedir, target_filepath)
            except IOError, value:
                message = """Target file %s is not writable
                on server %s (%s).\n""" % (
                    basedir + target_filepath, self.getHostname(), value)
                message.replace('\n','<br/>')                
                return MessageDialog(title   = 'ERROR',
                                     message = message,
                                     action  = 'create')
        message = """Object instance %s successfully created,
        pointing to file %s on %s.
        %s\n""" % (
            file_id, basedir + target_filepath, self.getHostname(), create_msg )

    self.create_object(file_id, file_title,
                       file_description, basedir,
                       target_filepath,
                       REQUEST, RESPONSE)

    message.replace('\n','<br/>')
    return MessageDialog(title   = 'CREATED',
                         message = message,
                         action  = 'manage_main')

def create_file(self, contents_file,
                basedir, target_filepath):
    """
    
    Creates a physical file on the filesystem at
    basedir+targetfilepath.  If Contents_file is not None, it points
    to a file whose contents will be copied into the targetfile.
    Otherwise, a zero length file will be created.  This method will
    raise and exception if there are any problems writing to the
    targetfile (or reading contents_file).

    """

    if contents_file:
        _copy(contents_file, basedir + target_filepath)
    else:
        touchfile = open(basedir + target_filepath, 'wb') # create zero-length file
        touchfile.close()
        
    return "File successfully created.\n"

def create_object(self, id, title,
                  description, basedir,
                  target_filepath,
                  REQUEST, RESPONSE):
    """

    Factory method to actually create the instance of ExternalFile.
    This method will be overridden in classes that inherit from
    ExternalFile.

    """
    self._setObject(id, ExternalFile(id, title,
                                     description,
                                     basedir + target_filepath))
    self._getOb(id).reindex_object()

def manage_add(self,
               id='', title='', description='',                   # standard
               target_filepath='', basedir='', upload_file=None,  # externalfile
               REQUEST=None, RESPONSE=None):
    """
    
    Workhorse method called from create.dtml form.  Actually does the
    work of confirming the user's intended (possibly destructive)
    actions, creating the target file if necessary, and calling
    create_object to actually create the object.  This method may be
    called from create callbacks of derived classes to do the heavy
    lifting.  Derived classes that call this method must define a
    create_object method (see example above) to actually create the
    proper instance.

    """
    success_url  = 'manage_main'
    tryagain_url = 'create'
    
    if not target_filepath:
        return MessageDialog(title   = 'ERROR',
                             message = 'You must specify a target filepath.',
                             action  = tryagain_url)
    if not id:
        id = target_filepath
        id = id[max(string.rfind(id,'/'), 
                    string.rfind(id,'\\'), 
                    string.rfind(id,':')
                    )+1:]

    # Six possible scenarios:
    #
    # [One]   No upload, target filepath already exists on the server and is readable.
    #         Action: Point to it.
    #
    # [Two]   No upload, target filepath does not exist on the server.
    #         Action: Give option to create blank file or cancel.
    #
    # [Three] No upload, target filepath exists on the server but is not readable.
    #         Action: Display error message, allow user to try again.
    #
    # [Four]  Upload, target filepath does not exist on the server.
    #         Action: Give option to create file with uploaded contents or cancel.
    #
    # [Five]  Upload, target filepath exists on the server but is not writeable.
    #         Action: Display error message, allow user to try again.
    #
    # [Six]   Upload, target filepath exists on the server and is writable.
    #         Action: Give options: overwrite/dont overwrite/try again
    #
    # [Seven] Upload, upload file is empty
    #         Action: Display error message, allow user to try again
    #
    if not upload_file or not upload_file.filename:
        if not os.path.exists(basedir + target_filepath):
            # Scenario Two
            message= """File %s does not exist yet on server %s.
            Would you like to create a blank file by that name?\n""" % ( basedir + target_filepath,
                                                                       self.getHostname() )
            return self.confirm_create_form(message          = message,
                                            file_id          = id, # id -> file_id or acquisition will get wrong one
                                            file_title       = title,
                                            file_description = description,
                                            basedir          = basedir,
                                            target_filepath  = target_filepath,
                                            temp_filename    = '',
                                            upload_filename  = '')
        else:
            try:
                testopen = open(basedir + target_filepath, 'rb')
            except IOError, value:
                # Scenario Three
                message = """Target file %s exists on server %s
                but is not readable (%s).\n""" % (
                    basedir + target_filepath, self.getHostname(), value )
                message.replace('\n','<br/>')
                return MessageDialog(title   = 'ERROR',
                                     message = message,
                                     action  = tryagain_url)
            else:
                # Scenario One
                self.create_object(id, title, description, basedir,
                                   target_filepath,
                                   REQUEST, RESPONSE)
                message = """ExternalFile \'%s\' successfully created,
                pointing to file %s on %s.\n""" % (
                    id, basedir + target_filepath, self.getHostname() )
                message.replace('\n','<br/>')                
                return MessageDialog(title   = 'CREATED',
                                     message = message,
                                     action  = success_url)
    else:
        #
        # we have an uploaded file.  Grab the contents and store in a
        # temporary file, copy it over to target_pathname iff the user
        # confirms this potentially destructive action
        #
        temp_filename = tempfile.mktemp()
        upload_contents = upload_file.read()
        if not upload_contents:
            message = """Uploaded file %s is empty.
            Please specify a non-empty file.\n""" % upload_file.filename
            message.replace('\n','<br/>')
            return MessageDialog(title   = 'ERROR',
                                 message = message,
                                 action  = tryagain_url)
        else:
            data = StringIO(upload_contents)
            _copy(data, temp_filename)
            if not os.path.exists(basedir + target_filepath):
                # Scenario Four
                message = """File %s does not exist yet on server %s.
                Would you like to create a file by that name containing the uploaded contents?
                """ % ( basedir + target_filepath, self.getHostname() )
                return self.confirm_create_form(message          = message,
                                                file_id          = id, # id -> file_id or acquisition will get wrong one
                                                file_title       = title,
                                                file_description = description,
                                                basedir          = basedir,
                                                target_filepath  = target_filepath,
                                                temp_filename    = temp_filename,
                                                upload_filename  = upload_file.filename)
            else:
                try:
                    testopen = open(basedir + target_filepath, 'ab')
                    testopen.close()
                except IOError, value:
                    # Scenario Five
                    message = """Target file %s is not writable
                    on server %s (%s).\n""" % (
                        basedir + target_filepath, self.getHostname(), value )
                    message.replace('\n','<br/>')                    
                    return MessageDialog(title   = 'ERROR',
                                         message = message,
                                         action  = tryagain_url)
                else:
                    # Scenario Six
                    return self.conflict_create_form(file_id          = id, # change id -> file_id else conflict
                                                     file_title       = title,
                                                     file_description = description,
                                                     basedir          = basedir,
                                                     target_filepath  = target_filepath,
                                                     temp_filename    = temp_filename,
                                                     upload_filename  = upload_file.filename)

def getHostname(self):
    return socket.gethostname()

################################################################
# ExternalFile class
################################################################

class ExternalFile(CatalogAware,    # ZCatalog support
                   PropertyManager, # Property support
                   SimpleItem):

    """
    ExternalFile encapsulates a file in the filesystem as a Zope object.
    """
    
    edit_externalfile_formpart    = DTMLFile('dtml/edit_externalfile_formpart', globals())



    meta_type = 'External File'  # This is the name Zope will use for the Product in
                                 # the "addProduct" list

    behave_like_list    = [ 'DTMLDocument', 'DTMLMethod', 'Image' ]

    # This tuple defines our properties
    _properties = (
        {'id':'title',          'type':'string',    'mode': 'w'},
        {'id':'description',    'type':'string',    'mode': 'w'},
        {'id':'filepath',       'type':'string',    'mode': ''},
        {'id':'content_type',   'type':'selection', 'mode': 'w', 'select_variable':'getMimetypelist', },
        {'id':'behave_like',    'type':'selection', 'mode': 'w', 'select_variable':'behave_like_list', },        
        )
    
    # This tuple defines a dictionary for each tab in the management interface
    # label = label of tab, action = url it links to
    manage_options = (
        {'label':'Edit', 'action': 'manage_editForm', 'help': ('ExternalFile', 'edit.stx') },
        {'label':'View', 'action': 'manage_view',     'help': ('ExternalFile', 'view.stx') },
        ) + OFS.PropertyManager.PropertyManager.manage_options + OFS.SimpleItem.SimpleItem.manage_options

    _security = ClassSecurityInfo()

    # set security for the object itself, e.g. if it is accessed in DTML code
    # This line is REQUIRED to allow access to External Files by the public (ariel DTDs, etc.)
    _security.declareObjectPublic()

    def __init__(self, id, title, description, filepath):
        """
        Initialize a new instance of ExternalFile
        """
        self.id          = id       # This is the OID translated by Zope into a URL
        self.title       = title
        self.description = description
        self.filepath    = filepath # This is the full pathname of the file
        self.__version__ = __version__

        # This next line should be module scope, but I don't know if we can guarantee
        # execution order for Zope products....?
        mimetypes.types_map.update(addl_fileext_to_mime_types_map)
        if filepath:
            self.content_type, encoding  = mimetypes.guess_type(filepath)
            if not self.content_type:
                self.content_type = default_content_type
        else:
            self.content_type = default_content_type

        #
        # By default, behave like a DTMLDocument, unless this is an image,
        # in which case, we need to behave like an Image.
        #
        if self.getFiletype() == 'image':
            self.behave_like = 'Image'
        else:
            self.behave_like = 'DTMLDocument'

    # call __call__() method on DTMLFile for our edit form, point at our externalFileEdit.dtml
    # globals() is standard python for passing the current namespace
    _security.declareProtected('Edit', 'manage_editForm')
    manage_editForm = DTMLFile('dtml/edit',globals())

    # call __call__() method on DTMLFile for our index_html, point at our View.dtml
    # globals() is standard python for passing the current namespace
    _security.declareProtected('View', 'manage_view')
    manage_view           = DTMLFile('dtml/view',globals())

    _security.declareProtected('Access contents information', 'index_html')
    # next line is not strictly necessary, Access contents info is Anonymous by default
    # as opposed to all other protections that get Manager Role by default
    _security.setPermissionDefault('Access contents information', ('Anonymous')) 
    def index_html(self, REQUEST, RESPONSE):
        """

        Render the document/method/Image appropriately.  Note that
        this method is called for ExternalFile when it is emulating a
        DTMLMethod, even though a DTMLMethod itself never calls
        index_html.

        """
        return self.__call__(None, REQUEST, RESPONSE)

    manage_header_message = DTMLFile('dtml/manage_header_message', globals())

    _security.declareProtected('Access contents information', '__str__')
    def __str__(self, REQUEST=None, RESPONSE=None):
        """get contents as a string"""

        return self.__call__(None, REQUEST, RESPONSE)
        
    _security.declareProtected('Access contents information', '__call__')
    def __call__(self, client=None, REQUEST=None, RESPONSE=None):
        """Render the document given a client object, REQUEST mapping,
        Response, and key word arguments."""

        if self.behave_like == 'Image':
            resultImage = OFS.Image.Image(self.id, self.title,
                                          self.getContents(REQUEST, RESPONSE),
                                          self.content_type)
            resultImage._p_mtime     = self.getModTime(REQUEST, RESPONSE)
            resultImage.absolute_url = self.absolute_url
            if REQUEST:
                return resultImage.index_html(REQUEST, RESPONSE)
            else:
                return resultImage.__str__()
            
        elif self.behave_like == 'DTMLDocument':
            me = self.getSelf()
            resultDocument = OFS.DTMLDocument.DTMLDocument(
                self.getContents(REQUEST, RESPONSE))

            resultDocument.id    = self.id
            resultDocument.title = self.title
            return resultDocument(me, REQUEST, RESPONSE)
        
        else: # DTMLMethod
            myContainer = self.aq_chain[1]
            resultDocument = OFS.DTMLMethod.DTMLMethod(
                self.getContents(REQUEST, RESPONSE))
            
            resultDocument.id    = self.id
            resultDocument.title = self.title
            return resultDocument.__of__(myContainer)(myContainer, REQUEST, RESPONSE)

    _security.declarePublic('__len__')
    def __len__(self): return 1

    _security.declarePublic('getMimetypelist')
    def getMimetypelist(self):
        """
        
        Return list of available mime types.  Used to populate
        selection list for mimetype property.
        
        """
        typelist = mimetypes.types_map.values()

        # return sorted, unique-ified list.
        # Code stolen from python FAQ answer 4.64
        typelist.sort()
        last = typelist[-1]
        for i in range(len(typelist)-2, -1, -1):
          if last==typelist[i]: del typelist[i]
          else: last=typelist[i]
        return typelist

    # innocuous methods should be declared public
    _security.declarePublic('getSelf')
    def getSelf(self):
        """Return this object.
        For use in DTML scripts"""
        return self.aq_chain[0]
    
    # protect these b/c a malicious hacker could get info about the filesystem
    _security.declareProtected('View',  'getFiletype')
    def getFiletype(self):
        """

        Returns 'ascii', 'image', or 'binary' according to the
        mime_type_map defaults to 'binary' if no match is found

        """
        if mime_type_map.has_key(self.content_type):
            return mime_type_map[self.content_type]
        else:
            return 'binary'

    # protect these b/c a malicious hacker could get info about the filesystem
    _security.declareProtected('View',  'getFilepath')
    def getFilepath(self, REQUEST=None, RESPONSE=None):
        return self.filepath

    _size_changes = {
        # directive: (colDelta, rowDelta)
        'Narrower': (-5,  0),
        'Wider':    (5,   0),
        'Taller':   (0,   5),
        'Shorter':  (0,  -5),
        }

    _minCols = 40
    _minRows = 5

    _security.declareProtected('Edit', 'manage_editAction')
    def manage_editAction(self, externalfile_content, SUBMIT,
                          edit_textarea_cols = '80',
                          edit_textarea_rows = '40',
                          upload_file = None,
                          REQUEST=None, RESPONSE=None):
        """
        
        Called in response to edit.dtml form submit, to save changes,
        cancel changes, replace the contents of the file with uploaded
        contents, or change the size of the edit textarea form
        control.
        
        """
        if SUBMIT == 'Save Changes':
            self.setContents(externalfile_content, REQUEST, RESPONSE)
            message = 'Content changed. (%s)' % DateTime().ISO()
            return self.manage_editForm(header_message=message)

        elif SUBMIT == 'Cancel Changes':
            message = 'Content changes cancelled.'
            return self.manage_editForm(externalfile_content=None,
                                        header_message=message)

        elif SUBMIT == 'Upload File':
            valid   = 0
            message = "Please specify a a non-empty file to upload."
            if upload_file and upload_file.filename:
                upload_contents = upload_file.read()
                if not upload_contents:
                    message = """Uploaded file %s is empty.
                    Please specify a non-empty file.\n""" % upload_file.filename
                else:
                    valid = 1;

            if valid:
                self.setContents(upload_contents, REQUEST, RESPONSE)
                message = 'Content changed. (%s)' % DateTime().ISO()
                return self.manage_editForm(externalfile_content=None,
                                            header_message=message)
            else:
                message.replace('\n','<br/>')
                return MessageDialog(title   = 'ERROR',
                                     message = message,
                                     action  = 'manage_editForm')

        else: # change size of textarea form control 
            colDelta, rowDelta = self._size_changes[SUBMIT]
            cols=max(self._minCols, string.atoi(edit_textarea_cols)+colDelta)
            rows=max(self._minRows, string.atoi(edit_textarea_rows)+rowDelta)

            e=(DateTime('GMT') + 365).rfc822()

            resp = REQUEST['RESPONSE'] 
            resp.setCookie('edit_textarea_cols',str(cols),path='/',expires=e)
            resp.setCookie('edit_textarea_rows',str(rows),path='/',expires=e)
        
            return self.manage_editForm(self, externalfile_content, REQUEST,
                                        edit_textarea_cols = cols,
                                        edit_textarea_rows = rows,
                                        header_message='Size changed.')

    # Saving the file back via FTP or webdav
    _security.declareProtected('Edit', 'PUT')
    def PUT(self, REQUEST, RESPONSE):
        """ Handle HTTP PUT requests """
        self.dav__init(REQUEST, RESPONSE)
        instream = REQUEST['BODYFILE']
        self.setContents(instream.read(), REQUEST, RESPONSE)
        RESPONSE.setStatus(204)
        return RESPONSE

    _security.declareProtected('View', 'getFileModTime')
    def getFileModTime(self, REQUEST=None, RESPONSE=None):
        """
        
        Return the last modification date of the external file to
        which we are referring.  NOTE: this value is returned as
        number of seconds since the epoch in UTC (see python time
        module)
        
        """
        path = self.getFilepath(REQUEST, RESPONSE)
        return os.path.getmtime(path)

    _security.declareProtected('View', 'getFileModTime')
    def getModTime(self, REQUEST=None, RESPONSE=None):
        """
        
        Return the last modification date of the external file to
        which we are referring, or of our metadata which is stored
        internally in the ZODB, whichever is more recent (greater).
        NOTE: this value is returned as number of seconds since the
        epoch in UTC (see python time module)
        
        """
        tempTime = self.getFileModTime(REQUEST, RESPONSE)
        if self._p_mtime > tempTime:
            return self._p_mtime
        else:
            return tempTime

    _security.declareProtected('View', 'getContents')
    def getContents(self, REQUEST=None, RESPONSE=None):
        """
        Return the contents of this external file as a string.
        
        """
        #
        # if file does not exist, should we throw an exception?
        # right now, we return the error message as if it
        # were the contents of the file.
        #
        # We need to do something better! cks 10/31/2001
        #
        #print "getContents of:",self.getFilepath(REQUEST, RESPONSE)
        try:
            data = StringIO()
            path = self.getFilepath(REQUEST, RESPONSE)
            _copy(path, data)
        except IOError, value:
            return value
        else:
            return data.getvalue()

    _security.declareProtected('Edit', 'setContents')
    def setContents(self, newcontents, REQUEST=None, RESPONSE=None):
        """
        
        Overwrite the previous contents of this file with the
        newcontents string.
        

        """
        data = StringIO(newcontents)
        _copy(data, self.getFilepath(REQUEST, RESPONSE))
        # update ZCatalog
        self.reindex_object()
        # No need to set dirty flag, Zope object itself has not actually changed
        # self._p_changed = 1  # dirty flag, triggers persistence to update the object

    #  FTP "get"
    _security.declareProtected('FTP access', 'manage_FTPget')
    manage_FTPget = getContents

    _security.declareProtected('Search ZCatalog', 'PrincipiaSearchSource')
    def PrincipiaSearchSource(self, REQUEST=None, RESPONSE=None):
        """
        Support for full text search via ZCatalog indexing
        """
        if self.getFiletype() == 'ascii':
            return self.getContents(REQUEST, RESPONSE)
        else:
            return ""

    def repair(self):
        """

        Repair this object.  This method is used for schema migration,
        when the class definition changes and existing instances of
        previous versions must be updated.

        """
        if not hasattr(self, 'behave_like'):
            self.behave_like = 'DTMLDocument'

# register security information
Globals.InitializeClass(ExternalFile)
